/* TODO 
	Percent complete would be better if it included portion of current segment completed.
	Erase previous IDL's if desired.
	Lock model prior to entry and release lock after copying info.
	1/2 amp If classes are not ascending, sort causes reordering - need to preserve it.

*/

#include "stdafx.h"

#include "TaskIdealize.h"
#include "HMM.h"

//----------------------------------------------------------------------
void COptionsIdealize :: ToBuffer( BYTE * buf ) { 
	sprintf( (char *)buf, "2 %3d %1d %3d %1d %1d %3d,%3d,%3d %c%c %2d %s", 
			min(999,m_iChannel), min(9,m_iDataMode), min(999,m_iModelSelected), 
			min(9,m_iEraseMode), min(9,m_iMethod), 
			GetRValue(m_crColor), GetGValue(m_crColor), GetBValue(m_crColor), 
			m_bStatistics?'Y':'N', m_bHistogram?'Y':'N', 
			m_cSetValues.Length(), m_cSetValues.ConstBuffer() );
	}

void COptionsIdealize :: FromBuffer( BYTE * buf ) { 
	int iVer;
	int iRed, iGrn, iBlu;
	char chStat='Y';
	char chHist='Y';
	sscanf( (char *)buf, "%1d %3d %1d %3d %1d %1d %3d,%3d,%3d %c%c", 
			&iVer, &m_iChannel, &m_iDataMode, &m_iModelSelected, 
			&m_iEraseMode, &m_iMethod, &iRed, &iGrn, &iBlu, 
			&chStat, &chHist );
	m_bStatistics = ( chStat=='Y' );
	m_bHistogram = ( chHist=='Y' );
	m_crColor = RGB(iRed,iGrn,iBlu);
	if( iVer>=2 ) { 
		int iLen = atoi_l( (char *)buf+31, 2 );
		m_cSetValues.Set( (char *)buf+34, iLen );
		}
	}

bool COptionsIdealize :: RegistryLoad( ) { 
	CString c = RegistryLoadString( "OPTIONSIDEALIZE" ); 
	if( c.Length()==0 ) 
		return false;
	FromBuffer( (BYTE *)c.ConstBuffer() );
	return true;
	}

bool COptionsIdealize :: RegistrySave() { 
	BYTE buf[BUF_OPTIDL_SIZE+64];
	ToBuffer( buf );
	return RegistrySaveString( "OPTIONSIDEALIZE", (char *)buf );
	}

//----------------------------------------------------------------------
void IdealizeHalfAmp( CVwTaskIdealize * pVwTask, CSelectionSet * pSet, bool bFromModel );
void IdealizeSKM( CVwTaskIdealize * pVwTask, CSelectionSet * pSet );

DWORD WINAPI TaskIdealize( LPVOID pParam ) { 
	CVwTaskIdealize * pVwTask = (CVwTaskIdealize *) pParam;
	CSelectionSet * pSet = &(pVwTask->m_oSelSet);

	if( pVwTask->m_pOptionsIdealize->m_iMethod==IDLMTHD_HALFAMP ) 
		IdealizeHalfAmp( pVwTask, pSet, true );
	else if( pVwTask->m_pOptionsIdealize->m_iMethod==IDLMTHD_SKM ) 
		IdealizeSKM( pVwTask, pSet );
	else if( pVwTask->m_pOptionsIdealize->m_iMethod==IDLMTHD_SETVALUES ) 
		IdealizeHalfAmp( pVwTask, pSet, false );

	pSet->m_pData->SetStatus(0);
	PostMessage( pVwTask->m_hDlgVwData, WM_DATA_CHANGED, 0, 0 );	
	
	PostMessage( (HWND)pVwTask->m_hDlg, WM_TASKCOMPLETE, pVwTask->m_bCancelled?1:0, 0 );
	return -1;
	}

//-----------------------------------------------------------------------------

bool SKM( int iPoints,				// # of data points
			  int * aiData,				// [iPoints] - observation data
			  int iStates,				// # of states
			  double * adPr,			// [iStates] initial probability
			  double ** adLogA,			// 'A' transition matrix 
			  int iBins,				// 
			  double ** adLogB,			// Gaussian Weight Factors, log values >= -1e-10
										// B[i][j] is used for likelihood of sample in bin j being of state i
			  int * aiState,			// [iPoints] output state index for each point
			  double * pdLogLikelihood	// output result log-likelihood
			  ) {
	int i, j; 
	
	/*
	#ifdef _DEBUG
	static bool g_bFirst=true;
	FILE * h;
	if( g_bFirst ) { 
		h = fopen( "C:\\SKM.txt", "a" );
		fprintf( h, "points=%d state%d bins=%d\n", iPoints, iStates, iBins );
		LogArray_i( h, "aiData", aiData, iPoints );
		LogArray_d( h, "adPr", adPr, iStates );
		LogArray_dd( h, "adLogA", adLogA, iStates, iStates );
		}
	#endif
	*/

	//-- Initial 
	TNT::Array2D<short> aiMaxPos( iStates, iPoints );		
	TNT::Array2D<double> adMax( 2, iStates );
	double * adEval = new double[iStates];
	int iMax=0;					// Determines which adMax[2] is current and prior.  Alternates each iteration.
	
	//-- Initialize with starting probability * observation likelihood for each state.
	for( i=0; i<iStates; i++){					
		adMax[iMax][i] = ( adPr[i]>0.0 ? log(adPr[i]) : -1.0e10 ) + adLogB[i][aiData[0]];	
		aiMaxPos[i][0] = 0;       
		}

	//-- for each point determine highest probability of [path to prior state]*[current sample observation]
	for( int iPoint=1; iPoint<iPoints; iPoint++) {
		for( j=0; j<iStates; j++) {
			//-- For each possibility of transitioning from prior state i
			// to new state j, compute product of prior likelihood * transtion(i->j) 
			// * observation likelihood
			for( int i=0; i<iStates; i++ )			
				adEval[i] = adMax[iMax][i]+adLogA[i][j]+adLogB[j][aiData[iPoint]];
			//-- select maximum likelihood to arrive at state j
			adMax[1-iMax][j] = adEval[ aiMaxPos[j][iPoint]=FindMaxPos( iStates, adEval ) ];
			}
		iMax=1-iMax;
		}
		
	//-- backtrack to recover the state sequence 
	for( i=0; i<iStates; i++)
		adEval[i] = adMax[1-iMax][i];             

	//-- max log likelihood position and value.
	*pdLogLikelihood = adEval[ aiState[iPoints-1] = FindMaxPos( iStates, adEval ) ];
	for( iPoint=iPoints-2; iPoint>=0; iPoint-- ) {
		i = aiState[iPoint+1];
		aiState[iPoint] = int(aiMaxPos[i][iPoint+1]);           
		}

	delete [] adEval;
	/*
	#ifdef _DEBUG
	if( g_bFirst ) { 
		fprintf( h, "--------Results\nLL=%f\n", *pdLogLikelihood );
		LogArray_i( h, "aiState", aiState, iPoints );
		fclose(h);
		g_bFirst=false;
		}
	#endif
	*/
	return true;
	}

//------------------------------------------------------------------------
void IdealizeSKM( CVwTaskIdealize * pVwTask, CSelectionSet * pSet ) {
	const int BINCOUNT=2048;
	int i, j;
	int iStates = pVwTask->m_pModel->m_aStates.Count();
	double dSampling = pSet->m_pData->m_dSampling;

	//----- Process Model 	
	//-- Start probabilities of model 
	double * adPr = new double[ iStates ];
	pVwTask->m_pModel->NormalizeStartProbabilities( adPr );
	
	//-- Obtain A Matrix, convert to ln A matrix
	TNT::Array2D<double> aQ = pVwTask->m_pModel->Q( );
	HMM::CModelMath<double> oHMM( aQ );
	oHMM.CalcEigen();
	oHMM.CalcA( dSampling );
	TNT::Array2D<double> aA( oHMM.m_aA );
	double * pd, ** adA = aA;
	for( i=0; i<iStates; i++)
		for( j=0; j<iStates; j++) {
			pd = &(adA[i][j]);			// note: could skip this temp var, the optimizer does it anyhow (?)	adA[i][j] = adA[i][j]>0 ? log(adA[i][j]) : -1.0e10 );
			*pd = *pd > 0.0 ? log(*pd) : -1.0e10;
			}
			
	//-- Create array of model amps, stddev's
	int iClass, iClasses = pVwTask->m_pModel->m_aClasses.Count();
	int iUsedClasses = pVwTask->m_pModel->UsedClassCount();
	double * adAmp=new double[iUsedClasses];	// unsorted !
	double * adStd=new double[iUsedClasses];
	int * aiCls=new int[iUsedClasses];
	pVwTask->m_pModel->GetAmpStdArray( iUsedClasses, aiCls, adAmp, adStd, false );		

	//-- Gauss weighting factors
	TNT::Array2D<double> adB( iStates, BINCOUNT );

	//-- State results 
	TNT::Array1D<int> aiStates(pSet->MaxSize());

	//----- Process data segments	
	//-- Scan ranges 
	double dLL;										// Log likelihood result from SKM()
	double dMin, dMax, dDelta;						// min, max data value and size of bins.
	double * pdData;								// pointer to data element
	CIdlList * pList;								// pointer to new Idealization created
	int iPoint, iPoints;							// # of points in current selection
	int iPreviousPt;								// during dwell construction, end pt of last dwell.
	int iRange=0, iRanges=pSet->m_aRanges.Count();	// # of selections
	int iPtCount=0, iTotalPoints=pSet->TotalSize();	// total points in all sections
	while( pSet->GetNext() ) { 
		iPoints = pSet->m_pRange->m_iMax-pSet->m_pRange->m_iMin+1;
		
		//-- Update the pct complete 
		PostMessage( pVwTask->m_hDlg, WM_TASKSETPCT, MulDiv(iPtCount+iPoints/2,100,iTotalPoints), 0 );
		iPtCount += iPoints;
		
		//-- create new IDL and copy amp's to it.
		pList = new CIdlList; 
		pList->m_crColor=pVwTask->m_pOptionsIdealize->m_crColor;
		pList->m_iChannel=pVwTask->m_pOptionsIdealize->m_iChannel;
		pList->m_iOffset=pSet->m_pRange->m_iMin;
		pList->m_iLength=iPoints;
		pList->AllocateClasses( iUsedClasses );
		for( iClass=0; iClass<iUsedClasses; ++iClass ) {
			pList->IdlClass(iClass)->m_dAmp=adAmp[iClass];
			pList->IdlClass(iClass)->m_dStdDev=adStd[iClass];
			}

		//-- dMin, dMax, dDelta : Calculate min and max values, and width of bins.
		FindRange( iPoints, pSet->m_adData, &dMin, &dMax );
		dDelta = (dMax-dMin)/BINCOUNT;

		//-- adB : Create the B gauss weight factors array and aiData map of adData
		for( i=0; i<iStates; ++i ) {
			double dAmp = pVwTask->m_pModel->StateAmp(i);
			double dStd = pVwTask->m_pModel->StateStd(i);
			for( j=0; j<BINCOUNT; ++j ) 
				adB[i][j] = GaussianWeightFactor( dAmp, dStd, dMin+dDelta*j );
			}
				
		//-- aiStates : find the bin corresponding to each data point
		if( dDelta==0.0 ) 
			dDelta=1.0;
		pdData=pSet->m_adData;
		for( iPoint=0; iPoint<iPoints; ++iPoint ) {
			aiStates[iPoint] = max(0, min(BINCOUNT-1, int((*pdData-dMin)/dDelta) ) );
			++pdData;
			}

		//-- Execute the SKM calculation
		SKM( iPoints, aiStates, iStates, adPr, adA, BINCOUNT, adB, aiStates, &dLL );
		
		//-- Convert aiStates array to dwells 
		iPreviousPt=0;
		for( i=1; i<=iPoints; ++i ) { 
			if( i==iPoints || aiStates[i]!=aiStates[i-1] )  { 
				pList->AddDwell( pVwTask->m_pModel->State(aiStates[i-1])->m_iClass, i-iPreviousPt, aiStates[i-1] );
				iPreviousPt=i;
				}
			}
			
		//-- Check for cancelled.
		if( pVwTask->m_bCancelled ) 
			break;
			
		//-- Update the idealization results - upgrade lock to write status , add IDL to list 
		pSet->m_pData->SetStatus(2);		//TODO - loop ? error on fail ? 
		pSet->m_pData->Segment(pSet->m_pRange->m_iSeg)->m_aIdl.AddItem( pList );
		pSet->m_pData->SetStatus(1);
		
		++iRange;
		}

	pSet->m_pData->SetStatus(0);
	//-- 
	delete [] aiCls;
	delete [] adStd;
	delete [] adAmp;
	delete [] adPr;
	}

//-----------------------------------------------------------------------------
void IdealizeHalfAmp( CVwTaskIdealize * pVwTask, CSelectionSet * pSet, bool bFromModel ) { 
	//----- Process model, Create array of amps,stddev's
	int iClass=0, iUsedClasses;	//iClasses = pVwTask->m_pModel->m_aClasses.Count();

	if( bFromModel ) { 
		iUsedClasses = pVwTask->m_pModel->UsedClassCount();
		}
	else { 
		iUsedClasses = 0;
		int iLen = pVwTask->m_pOptionsIdealize->m_cSetValues.Length();
		char * pch = pVwTask->m_pOptionsIdealize->m_cSetValues.ConstBuffer();
		for( int i=0; i<iLen; ++i ) 
			if( pch[i]==',' ) 
				++iUsedClasses;
		}
	
	double * adAmp=new double[iUsedClasses];
	double * adAmpMid=new double[iUsedClasses];		// actually 1 extra
	double * adStd=new double[iUsedClasses];		// not really used 
	int * aiCls=new int[iUsedClasses];				// not really used 
	
	if( bFromModel ) { 
		pVwTask->m_pModel->GetAmpStdArray( iUsedClasses, aiCls, adAmp, adStd, true ); 
		//TODO - Release model lock here
		}
	else { 
		int iLen = pVwTask->m_pOptionsIdealize->m_cSetValues.Length();
		char * pch = pVwTask->m_pOptionsIdealize->m_cSetValues.ConstBuffer();
		int iPos=0;
		for( int i=0; i<=iLen && iClass<iUsedClasses; ++i ) 
			if( i==iLen || pch[i]==','  ) {
				adAmp[iClass]=atof_l( pch+iPos, i-iPos );
				adStd[iClass]=0.0;
				++iClass;
				}
				
		//-- slow sort, but list is small and it is fast if amps are already sorted(likely).
		iClass=1;
		while( iClass< iUsedClasses ) { 
			if( adAmp[iClass]<adAmp[iClass-1] ) { 
				Swap( adAmp+iClass, adAmp+(iClass-1) );
				if( iClass>0 ) 
					iClass--;
				}
			else 
				++iClass;
			}
		}
	
	//-- Find amp midpoints 
	for( iClass=0; iClass<iUsedClasses-1; ++iClass )
		adAmpMid[iClass]=(adAmp[iClass]+adAmp[iClass+1])/2.0;

	//-- Scan ranges 
	CIdlList * pList;
	double dPoint;
	int iPreviousPt, iPreviousClass, iPoint, iPoints;
	int iRange=0, iRanges=pSet->m_aRanges.Count();

//TODO	bool bHistogram = pVwTask->m_pOptionsIdealize->m_bHistogram;

	while( pSet->GetNext() ) { 

		//-- Update the pct complete 
		PostMessage( pVwTask->m_hDlg, WM_TASKSETPCT, MulDiv(++iRange,100,iRanges), 0 );
		
		//-- create new IDL and copy amp's to it.
		pList = new CIdlList; 
		pList->m_crColor=pVwTask->m_pOptionsIdealize->m_crColor;
		pList->m_iChannel=pVwTask->m_pOptionsIdealize->m_iChannel;
		pList->m_iOffset=pSet->m_pRange->m_iMin;
		pList->m_iLength=pSet->m_pRange->m_iMax-pSet->m_pRange->m_iMin+1;
		pList->AllocateClasses( iUsedClasses );
		for( iClass=0; iClass<iUsedClasses; ++iClass ) {
			pList->IdlClass(iClass)->m_dAmp=adAmp[iClass];
			pList->IdlClass(iClass)->m_dStdDev=adStd[iClass];
			}

		//-- Scan the points - add in dwells	
		iPreviousPt=0;
		iPreviousClass=-1;
		iPoints=pList->m_iLength;
		for( iPoint=0; iPoint<iPoints; ++iPoint ) { 
			//-- get next data 
			dPoint=pSet->m_adData[iPoint];
			
			//-- determine class of this point [iClass]
			for( iClass=0; iClass<iUsedClasses-1; ++iClass ) 
				if( adAmpMid[iClass]>=dPoint ) 
					break;
					
			//--if class changes, add a dwell
			if( iClass!=iPreviousClass || iPoint==iPoints-1 ) { 
				if( iPoint>0 )
					pList->AddDwell( iPreviousClass, iPoint-iPreviousPt );

				iPreviousClass=iClass;
				iPreviousPt=iPoint;
				}

			//-- check if cancelled 
			if( pVwTask->m_bCancelled ) 
				break;
			}
				
		//-- Update the idealization results - upgrade lock to write status , add IDL to list 
		if( ! pVwTask->m_bCancelled ) { 
			pSet->m_pData->SetStatus(2);		//TODO - loop ? error on fail ? 
			pSet->m_pData->Segment(pSet->m_pRange->m_iSeg)->m_aIdl.AddItem( pList );
			pSet->m_pData->SetStatus(1);
			}
			
		}
	
	pSet->m_pData->SetStatus(0);

	//-- Release the data 
	delete [] aiCls;
	delete [] adAmpMid;
	delete [] adAmp;
	delete [] adStd;
	}


